#
#   PROBE - Rocket propulsion and orbital mechanics calculations
#

from math import log, exp, sqrt
from constants import G

def delta_v(Isp, m0, dm):
	#  Change in velocity from burning a mass dm of fuel
	#  starting with a total vehicle mass of m0.
	return Isp * log(m0 / (m0 - dm))

def fuel_for_delta_v(Isp, m1, delta_v):
	#  Mass of fuel needed to change velocity by delta_v
	#  given a final vehicle mass of m1.
	m0 = m1 * exp(delta_v / Isp)
	return m0 - m1

def escape_velocity(M, r):
	#  Speed required to escape from a body of mass M
	#  starting at a distance r from its centre.
	return sqrt(2 * G * M / r)

def gravity(M, r):
	#  Gravitational acceleration at a distance r from
	#  the centre of a body of mass M
	return G * M / (r * r)

#def fuel_for_liftoff(Isp, b, m, g, v, e):
#	#  Calculate fuel required to boost a payload of
#	#  mass m to a velocity v in the presence of a
#	#  gravitational acceleration g (assumed constant).
#	#  b = burn rate
#	#  e = required relative precision of result
#	print "rocket_science.fuel_for_liftoff:" ###
#	print "  Isp = %.3e Ns/kg" % Isp ###
#	print "  b   = %.3e kg/s" % b ###
#	print "  m   = %.3e kg" % m ###
#	print "  g   = %.3e m/s^2" % g ###
#	print "  v   = %.3e m/s" % v ###
#	print "  e   = %.3e" % e ###
#	fv = fuel_for_delta_v(Isp, m, v)
#	print "Fuel to reach v = %.3e kg" % fv ###
#	tb = fv / b
#	print "Burn time = %.3f s" % tb ###
#	v1 = v
#	n = 5
#	while n:
#		n -= 1
#		vg = g * tb
#		print "Gravitational delta_v = %.3e m/s" % vg ###
#		fg = fuel_for_delta_v(Isp, m + 0.5 * fv, vg)
#		print "Estimated additional fuel = %.3e kg" % fg ###
#		tb2 = (fv + fg) / b
#		print "Revised burn time = %.3e s" % tb2 ###
#		vg2 = g * tb2
#		v2 = delta_v(Isp, m + fv + fg, fv + fg) - vg2
#		print "Revised final velocity = %.3e m/s" % v2 ###
#		d = abs((v1 - v2) / v1)
#		print "Relative error = %.2f" % d
#		if d <= e:
#			break
#		tb = tb2
#		fv += fg
#		v1 = v2
#	return fv + fg

def newton_raphson(f, fdash, x0, e):
	debug = False
	if debug:
		print "Newton-Raphson:" ###
		print "   x0 = %.3e" % x0 ###
	n = 10
	while n:
		n -= 1
		f_x0 = f(x0)
		fdash_x0 = fdash(x0)
		if debug:
			print "   f(x0) = %.3e" % f_x0
			print "   f'(x0) = %.3e" % fdash_x0
		x1 = x0 - f_x0 / fdash_x0
		if debug:
			print "   x1 = %.3e" % x1 ###
		d = abs((x1 - x0) / x0)
		if debug:
			print "   d = %.3e" % d
		if d <= e:
			return x0
		x0 = x1
	raise ValueError("Convergence failure")

def fuel_for_liftoff(Isp, b, m, g, v, e):
	#  Calculate fuel required to boost a payload of
	#  mass m to a velocity v in the presence of a
	#  gravitational acceleration g (assumed constant).
	#  b = burn rate
	#  e = required relative precision of result
	debug = False
	if debug:
		print "rocket_science.fuel_for_liftoff:" ###
		print "  Isp = %.3e Ns/kg" % Isp ###
		print "  b   = %.3e kg/s" % b ###
		print "  m   = %.3e kg" % m ###
		print "  g   = %.3e m/s^2" % g ###
		print "  v   = %.3e m/s" % v ###
		print "  e   = %.3e" % e ###
	fv = fuel_for_delta_v(Isp, m, v)
	if debug:
		print "Estimated fuel to reach v = %.3e kg" % fv ###
		print "Estimated burn time = %.3e" % (fv / b) ###
	def f(m1):
		#print "f(m1): m = %.3e, m1 = %.3e" % (m, m1) ###
		vf = delta_v(Isp, m1, m1 - m)
		vg = g * (m1 - m) / b
		if debug:
			print "      vf = %.3e" % vf ###
			print "      vg = %.3e" % vg ###
		return vf - vg - v
	def fdash(m1):
		return Isp / m1 - g / b
	m1 = m + fv
	if debug:
		print "Est slope = %.3e - %.3e" % (delta_v(Isp, m1, 1), g / b) ###
	m2 = newton_raphson(f, fdash, m1, e)
	if debug:
		print "Done newton_raphson"
	f = m2 - m
	tb = f / b
	vf2 = delta_v(Isp, m + f, f)
	vg2 = g * tb
	v2 = vf2 - vg2
	if debug:
		print "Revised fuel = %.3e" % f ###
		print "Burn time = %.3e" % tb ###
		print "Final vf = %.3e" % vf2 ###
		print "Final vg = %.3e" % vg2 ###
		print "Final velocity = %.3e" % v2 ###
	return f
