#
#   PROBE - Star systems
#

from math import exp, sqrt
from random import randint, uniform, normalvariate
from objects import Star, TerrestrialPlanet, GasGiantPlanet, RockyPlanet, \
	system_radius
from names import random_word
from constants import G
from vector_algebra import add, subtract, multiply, dot

class System(object):
	#  name       string
	#  star       Star
	#  home       TerrestrialPlanet
	#  objects    [SpaceObject]
	#  selection  SpaceObject
	#  vehicle    Vehicle
	
	selection = None
	vehicle = None

	def __init__(self):
		self.objects = []
		self.name = random_system_name()
		self.star = Star(self.name)
		self.objects.append(self.star)
		num_planets = random_num_planets()
		home_planet_number = randint(0, num_planets // 2)
		band = system_radius / (num_planets + 1)
		for i in xrange(num_planets):
			r = (i + 1.5) * band + normalvariate(0, 0.25 * band)
			if i == home_planet_number:
				planet = TerrestrialPlanet(r)
				self.home = planet
			else:
				p = gaussian(r, system_radius * 2 / 3, system_radius / 9)
				if uniform(0, 1) < p:
					planet = GasGiantPlanet(r)
				else:
					planet = RockyPlanet(r)
			self.objects.append(planet)
	
	def select_object(self, obj):
		#print "System: Selected", obj
		self.selection = obj
	
	def get_astronomical_data(self, data):
		for obj in self.objects:
			obj.get_astronomical_data(data)
			data.append(("", ""))
	
	def end_countdown(self):
		vehicle = self.vehicle
		if vehicle and vehicle.state == vehicle.PRELAUNCH:
			vehicle.state = vehicle.FLYING
	
	def timestep(self, game, dt):
		#print "System.timestep:", dt ###
		vehicle = self.vehicle
		if vehicle:
			vehicle.timestep(game, dt)
			if vehicle.state == vehicle.FLYING:
				self.apply_gravity(game, vehicle, dt)
	
	def apply_gravity(self, game, vehicle, dt):
		#print "System.apply_gravity:" ###
		p = vehicle.position
		asum = (0.0, 0.0)
		dists = []
		for obj in self.objects:
			if obj is not vehicle:
				M = obj.mass
				u = subtract(obj.position, p)
				rsq = dot(u, u)
				r = sqrt(rsq)
				a = multiply(G * M / (rsq * r), u)
				asum = add(asum, a)
				dists.append((r, obj))
				#if obj is self.home: ###
					#print "r = %.3e, a = (%.3e, %.3e)" % (r, a[0], a[1]) ###
		vehicle.velocity = add(vehicle.velocity, multiply(dt, asum))
		rmin, nearest = min(dists)
		vehicle.nearest_to = nearest
		vehicle.make_measurements(game, nearest, rmin)
		if rmin < nearest.radius - 10.0:
			print "CRASH!!!"
			vehicle.state = vehicle.CRASHED
	
	def set_vehicle(self, vehicle):
		if self.vehicle:
			self.objects.remove(self.vehicle)
		self.vehicle = vehicle
		if vehicle:
			self.objects.append(vehicle)


def random_system_name():
	#return "Betelgeuse"
	n = randint(1, 3) + randint(1, 3) + randint(1, 3)
	return random_word(n)

def random_num_planets():
	return randint(1, 3) + randint(1, 3) + randint(1, 3)

def gaussian(x, mean, sd):
	x2 = (x - mean) / sd
	return exp(-x2 * x2)

if __name__ == "__main__":
	system = System()
	print "System name:", system.name
	print "Star name:", system.star.name
	print "Home planet:", system.home.name
	for obj in system.objects:
		print
		obj.dump()
