#
#   PROBE - System Display
#

from pygame import draw
from pygame.locals import *
from albow.widget import Widget
from albow.resource import get_font, get_image
import objects
import vehicle
from objects import system_radius
from vector_algebra import distance

class ScreenSymbology:

	#bg_image = None
	#bg_color = None
	outline_symbols = False
	label_color = None
	selection_color = (128, 128, 255)
	orbit_color = (128, 128, 128)
	scale_color = (255, 255, 255)

	symbol_radii = {
		objects.Star: 6,
		objects.GasGiantPlanet: 4,
		objects.RockyPlanet: 2,
		objects.TerrestrialPlanet: 2,
		vehicle.Vehicle: 1,
	}
	
	symbol_colors = {
		objects.Star: (255, 255, 0),
		objects.GasGiantPlanet: (255, 255, 255),
		objects.RockyPlanet: (255, 255, 255),
		objects.TerrestrialPlanet: (0, 255, 0),
		vehicle.Vehicle: (255, 0, 0),
	}

class PaperSymbology:

	#bg_image = get_image("square_paper.jpg")
	#bg_color = (255, 255, 255)
	outline_symbols = True
	label_color = (0, 0, 0)
	selection_color = (0, 0, 255)
	orbit_color = (128, 128, 128)
	scale_color = (0, 0, 0)

	symbol_radii = {
		objects.Star: 6,
		objects.GasGiantPlanet: 5,
		objects.RockyPlanet: 4,
		objects.TerrestrialPlanet: 4,
		vehicle.Vehicle: 2,
	}
	
	symbol_colors = {
		objects.Star: (255, 255, 0),
		objects.GasGiantPlanet: (0, 0, 0),
		objects.RockyPlanet: (0, 0, 0),
		objects.TerrestrialPlanet: (0, 255, 0),
		vehicle.Vehicle: (255, 0, 0),
	}

class SystemDisplay(Widget):
	#  system   System
	#  origin   World coords of display centre
	#  zoom     Zoom factor (>= 1)

	font = get_font(10, "Vera.ttf")
	scale_font = get_font(10, "Vera.ttf")
	max_zoom = 1048576

	def __init__(self, rect, game, paper):
		Widget.__init__(self, rect)
		self.game = game
		if paper:
			self.symbology = PaperSymbology
		else:
			self.symbology = ScreenSymbology
		#self.bg_color = self.symbology.bg_color
		self.reset_zoom()
	
	def zoom_in(self):
		system = self.game.current_system
		if system:
			target = system.selection
			if target:
				self.origin = target.position
			if self.zoom < self.max_zoom:
				self.zoom *= 2
	
	def zoom_out(self):
		if self.zoom > 1:
			self.zoom /= 2
	
	def auto_zoom(self):
		system = self.game.current_system
		if system:
			target = system.selection
			vehicle = system.vehicle
			if target and vehicle:
				r = distance(vehicle.position, target.position)
				s = min(self.rect.width, self.rect.height) / 4
				self.reset_zoom()
				while self.zoom < self.max_zoom and self.world_to_screen_dist(r) < s:
					self.zoom_in()
	
	def reset_zoom(self):
		self.origin = (0.0, 0.0)
		self.zoom = 1
	
	def handle_key(self, e):
		key = e.key
		if key == K_UP:
			if e.shift:
				self.auto_zoom()
			else:
				self.zoom_in()
		elif key == K_DOWN:
			if e.shift:
				self.reset_zoom()
			else:
				self.zoom_out()
	
	def draw(self, surf):
		frame = surf.get_rect()
		#bg = self.symbology.bg_image
		#if bg:
		#	surf.blit(bg, (0, 0))
		system = self.game.current_system
		if system:
			if self.zoom <= 8:
				self.draw_orbits(surf, system)
			self.draw_objects(surf, system)
			self.draw_selection(surf, system)
		self.draw_scale(surf)
	
	def draw_scale(self, surf):
		font = self.scale_font
		color = self.symbology.scale_color
		min_sd = self.rect.width / 10 - 20
		sf = self.world_to_screen_scale()
		wd = 1.0
		while sf * wd < min_sd:
			wd *= 10
		sd = int(sf * wd)
		left = 40
		right = left + sd
		y = 20
		t = 5
		draw.line(surf, color, (left, y), (right, y), 1)
		draw.line(surf, color, (left, y - 2), (left, y + 2), 1)
		draw.line(surf, color, (right, y - 2), (right, y + 2), 1)
		buf = font.render("%.2e km" % (1e-3 * wd), True, color)
		r = buf.get_rect()
		r.midbottom = (left + sd / 2, y - 5)
		surf.blit(buf, r)
	
	def draw_orbits(self, surf, system):
		color = self.symbology.orbit_color
		if color:
			#c = self.center()
			c = self.world_to_screen_pos((0.0, 0.0))
			for obj in system.objects:
				r_orb = obj.orbital_distance
				if r_orb is not None:
					r = self.world_to_screen_dist(obj.orbital_distance)
					if r:
						draw.circle(surf, color, c, r, 1)
	
	def draw_objects(self, surf, system):
		#print "SystemDisplay.draw_objects: zoom factor =", self.zoom ###
		bl, bt = 0, 0
		br, bb = self.rect.size
		for obj in system.objects:
			if not (isinstance(obj, vehicle.Vehicle) and obj.state == obj.UNPREPPED):
				world_pos = obj.position
				screen_pos = self.world_to_screen_pos(world_pos)
				r = self.symbol_radius(obj)
				if r <= 2048:
					#if obj is system.star: ###
					#	print "SystemDisplay.draw_objects: Star symbol radius =", r ###
					sx, sy = screen_pos
					if sx + r > bl and sx - r < br and sy + r > bt and sy - r < bb:
						color = self.symbol_color(obj)
						draw.circle(surf, color, screen_pos, r)
						if self.symbology.outline_symbols:
							draw.circle(surf, (0, 0, 0), screen_pos, r, 1)
						self.draw_label(surf, obj, screen_pos, color, r)
	
	def draw_label(self, surf, obj, screen_pos, color, d):
		frame = surf.get_rect()
		x, y = screen_pos
		d = max(d, 3)
		color = self.symbology.label_color or color
		buf = self.font.render(obj.name, True, color)
		r = buf.get_rect()
		dx, dy = obj.position
		if isinstance(obj, vehicle.Vehicle):
			dy = -dy
		if dx < 0:
			if dy < 0:
				r.bottomright = (x - d, y - d)
			else:
				r.topright = (x - d, y + d)
		else:
			if dy < 0:
				r.bottomleft = (x + d, y - d)
			else:
				r.topleft = (x + d, y + d)
		if r.left < frame.left:
			r.left = frame.left
		if r.right > frame.right:
			r.right = frame.right
		surf.blit(buf, r)
	
	def draw_selection(self, surf, system):
		obj = system.selection
		if obj:
			p = self.world_to_screen_pos(obj.position)
			r = self.symbol_radius(obj)
			r1 = r + 3
			r2 = r1 + 10
			cx, cy = self.world_to_screen_pos(obj.position)
			color = self.symbology.selection_color
			draw.circle(surf, color, p, r1, 1)
			draw.line(surf, color, (cx - r1, cy), (cx - r2, cy))
			draw.line(surf, color, (cx + r1, cy), (cx + r2, cy))
			draw.line(surf, color, (cx, cy - r1), (cx, cy - r2))
			draw.line(surf, color, (cx, cy + r1), (cx, cy + r2))

	def mouse_down(self, e):
		system = self.game.current_system
		if system:
			hits = []
			m = e.local
			for obj in system.objects:
				p = self.world_to_screen_pos(obj.position)
				d = distance(m, p)
				r = max(3, self.symbol_radius(obj))
				if d <= r:
					hits.append((d, obj))
			if hits:
				d, obj = min(hits)
				system.select_object(obj)
	
	def world_to_screen_pos(self, p):
		cx, cy = self.center()
		s = self.world_to_screen_scale()
		ox, oy = self.origin
		x, y = p
		return int(round(cx + s * (x - ox))), int(round(cy + s * (y - oy)))
	
	def world_to_screen_dist(self, d):
		s = self.world_to_screen_scale()
		return int(round(s * d))
	
	def world_to_screen_scale(self):
		width, height = self.rect.size
		return self.zoom * min(width, height) * 0.5 / system_radius
	
	def center(self):
		#  Centre of display in pixels
		width, height = self.rect.size
		return width // 2, height // 2

	def symbol_radius(self, obj):
		r1 = self.world_to_screen_dist(obj.radius)
		r2 = self.symbology.symbol_radii.get(obj.__class__, r1)
		return max(r1, r2)

	def symbol_color(self, obj):
		return self.symbology.symbol_colors[obj.__class__]
